import mxnet as mx
from ...common.exceptions import InferenceError
from ..variables import VariableType
from .function_evaluation import FunctionEvaluation, FunctionEvaluationDecorator


class GluonFunctionEvaluation(FunctionEvaluation):
    """
    The evaluation of a function that is a wrapper of a MXNet Gluon block.

    :param function_wrapper: the MXFusion wrapper of the MXNet Gluon block that the function evaluation is associated with.
    :type function_wrapper: MXFusion.functions.MXFusionGluonFunction
    :param input_variables: the input arguments to the function.
    :type input_variables: {variable name : Variable}
    :param block_variables: the parameters in the Gluon block.
    :type block_variables: {variable name : Variable}
    :param num_outputs: the number of outputs of the function.
    :type num_outputs: int
    :param broadcastable: Whether the function supports broadcasting with the additional dimension for samples.
    :type: boolean
    """

    def __init__(self, function_wrapper, input_variables, block_variables,
                 output_variables, broadcastable=False):
        self.function_wrapper = function_wrapper
        self._input_variables_names = [k for k, _ in input_variables]
        self.block_variables_names = [k for k, _ in block_variables]

        inputs = block_variables + input_variables

        for _, bv in block_variables:
            if bv.type == VariableType.RANDVAR and broadcastable:
                # Broadcasting function evaluation can not be applied to the Gluon block with gluon block parameters as random variables.
                broadcastable = False

        input_names = [k for k, _ in inputs]
        output_names = [k for k, _ in output_variables]

        super(GluonFunctionEvaluation, self).__init__(
            inputs, output_variables,
            input_names, output_names, broadcastable=broadcastable)

    @property
    def _input_variables(self):
        return [(k, v) for k, v in self.inputs if k in self._input_variables_names]

    @property
    def block_variables(self):
        return [(k, v) for k, v in self.inputs if k in self.block_variables_names]

    def replicate_self(self, attribute_map=None):
        replicant = super(GluonFunctionEvaluation, self).replicate_self(attribute_map)
        replicant.function_wrapper = self.function_wrapper
        replicant.block_variables_names = self.block_variables_names
        replicant._input_variables_names = self._input_variables_names
        return replicant

    @FunctionEvaluationDecorator()
    def eval(self, F, **input_kws):
        """
        Invokes the MXNet Gluon block with the arguments passed in.

        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
        :param \**input_kws: the dict of inputs to the functions. The key in the dict should match with the name of inputs specified in the inputs
            of FunctionEvaluation.
        :type \**input_kws: {variable name: MXNet NDArray or MXNet Symbol}
        :returns: the return value of the function
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        inputs_func = [input_kws[k] for k, _ in self._input_variables]
        self._override_block_parameters(input_kws)
        return self.function_wrapper.block(*inputs_func)

    def _override_block_parameters(self, input_kws):
        """
        When a probabilistic distribution is defined for the parameters of a Gluon block (in ParameterDict), a special treatment is necessary
        because otherwise these parameters will be directly exposed to a gradient optimizer as free parameters.

        For each parameters of the Gluon bock with probabilistic distribution, this method dynamically sets its values as the outcome of
        upstream computation and ensure the correct gradient can be estimated via automatic differenciation.

        :param \**input_kws: the dict of inputs to the functions. The key in the dict should match with the name of inputs specified in the
            inputs of FunctionEvaluation.
        :type \**input_kws: {variable name: MXNet NDArray or MXNet Symbol}
        """
        for bn, bv in self.block_variables:
            if bv.type == VariableType.RANDVAR:
                if self.broadcastable:
                    raise InferenceError('Broadcasting function evaluation can not be applied to the Gluon block with gluon block parameters as random variables.')
                val = input_kws[bn]
                param = self.function_wrapper.block.collect_params()[bn]

                if isinstance(val, mx.ndarray.ndarray.NDArray):
                    ctx = val.context
                    ctx_list = param._ctx_map[ctx.device_typeid&1]
                    if ctx.device_id >= len(ctx_list) or ctx_list[ctx.device_id] is None:
                        raise Exception
                    dev_id = ctx_list[ctx.device_id]
                    param._data[dev_id] = val
                else:
                    param._var = val



from .stationary import StationaryKernel
from .....util.customop import broadcast_to_w_samples


class RBF(StationaryKernel):
    """
    Radial Basis Function kernel, aka squared-exponential, exponentiated quadratic or Gaussian kernel:

    .. math::
       k(r^2) = \\sigma^2 \\exp \\bigg(- \\frac{1}{2} r^2 \\bigg)

    :param input_dim: the number of dimensions of the kernel. (The total number of active dimensions)
    :type input_dim: int
    :param ARD: a binary switch for Automatic Relevance Determination (ARD). If true, the squared distance is divided by a lengthscale for individual
        dimensions.
    :type ARD: boolean
    :param variance: the initial value for the variance parameter (scalar), which scales the whole covariance matrix.
    :type variance: float or MXNet NDArray
    :param lengthscale: the initial value for the lengthscale parameter.
    :type lengthscale: float or MXNet NDArray
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param active_dims: The dimensions of the inputs that are taken for the covariance matrix computation. (default: None, taking all the dimensions).
    :type active_dims: [int] or None
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, input_dim, ARD=False, variance=1., lengthscale=1.,
                 name='rbf', active_dims=None, dtype=None, ctx=None):
        super(RBF, self).__init__(input_dim=input_dim, name=name,
                                  active_dims=active_dims, dtype=dtype,
                                  ctx=ctx)

    def _compute_K(self, F, X, lengthscale, variance, X2=None):
        """
        The internal interface for the actual covariance matrix computation.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :param variance: the variance parameter (scalar), which scales the whole covariance matrix.
        :type variance: MXNet NDArray or MXNet Symbol
        :param lengthscale: the lengthscale parameter.
        :type lengthscale: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        R2 = self._compute_R2(F, X, lengthscale, variance, X2=X2)
        return F.exp(R2 / -2) * broadcast_to_w_samples(F, variance, R2.shape)



from .factor_graph import FactorGraph


class Posterior(FactorGraph):
    """
    A Posterior graph defined over an existing model.
    """

    def __init__(self, model, name=None):
        """
        Constructor.

        :param model:  The model which the posterior graph is defined over.
        :type model: Model
        """
        super(Posterior, self).__init__(name=name)
        self._model = model

    def __getattr__(self, name):
        if hasattr(self._model, name):
            v = getattr(self._model, name)

            replicant = v.replicate()
            setattr(self, name, replicant)
            return replicant
        else:
            raise AttributeError("''%s' object has no attribute '%s'" % (type(self), name))

    def __getitem__(self, item):
        if item in self.components:
            return self.components[item]
        elif item in self._model:
            object = self._model[item]
            replicant = object.replicate()
            if object.name is not None:
                setattr(self, object.name, replicant)
            else:
                replicant.graph = self.components_graph
            return replicant
        else:
            raise AttributeError("''%s' object has no item '%s'" % (type(self), item))

    def _replicate_class(self, **kwargs):
        """
        Return a new instance of the derived FactorGraph's class.
        """
        return Posterior(**kwargs)



from .inference_alg import InferenceAlgorithm


class StochasticVariationalInference(InferenceAlgorithm):
    """
    The class of the Stochastic Variational Inference (SVI) algorithm.

    :param num_samples: the number of samples used in estimating the variational lower bound
    :type num_samples: int
    :param model: the definition of the probabilistic model
    :type model: Model
    :param posterior: the definition of the variational posterior of the probabilistic model
    :param posterior: Posterior
    :param observed: A list of observed variables
    :type observed: [Variable]
    """

    def __init__(self, num_samples, model, posterior, observed):
        super(StochasticVariationalInference, self).__init__(
            model=model, observed=observed, extra_graphs=[posterior])
        self.num_samples = num_samples

    @property
    def posterior(self):
        """
        return the variational posterior.
        """
        return self._extra_graphs[0]

    def compute(self, F, data, parameters, constants):
        """
        The method for the computation of the inference algorithm

        :param F: the execution context (mxnet.ndarray or mxnet.symbol)
        :type F: Python module
        :param data: the data variables for inference
        :type data: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param parameters: the parameters for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :param constants: the constants for inference
        :type parameters: {Variable: mxnet.ndarray.ndarray.NDArray or
            mxnet.symbol.symbol.Symbol}
        :returns: the outcome of the inference algorithm
        :rtype: mxnet.ndarray.ndarray.NDArray or mxnet.symbol.symbol.Symbol
        """
        knowns = data.copy()
        knowns.update(parameters)
        knowns.update(constants)

        samples = self.posterior.draw_samples(
            F=F, conditionals=knowns, num_samples=self.num_samples,
            constants=constants)
        knowns.update(samples)
        logL = self.model.compute_log_prob(
            F=F, targets=knowns, constants=constants)
        logL = logL - self.posterior.compute_log_prob(
            F=F, targets=knowns, constants=constants)
        return -logL



"""This module contains Distributions for MXFusion.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    categorical
    distribution
    normal
    pointmass
    random_gen
    univariate
    gp
"""

__all__ = ['categorical', 'distribution', 'normal', 'pointmass', 'rand_gen',
           'univariate','gp']

from .distribution import Distribution
from .normal import Normal, MultivariateNormal
from .pointmass import PointMass
from .categorical import Categorical
from .gp import GaussianProcess, ConditionalGaussianProcess



"""This module contains utlity functions used throughout MXFusion.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    customop
    graph_serialization
    inference
    testutils
    util
"""



import numpy as np
import mxnet as mx
from ...common.config import get_default_MXNet_mode
from ..variables import Variable
from .univariate import UnivariateDistribution, UnivariateLogPDFDecorator, UnivariateDrawSamplesDecorator
from .distribution import Distribution, LogPDFDecorator, DrawSamplesDecorator
from ..variables import is_sampled_array, get_num_samples
from ...util.customop import broadcast_to_w_samples


class Normal(UnivariateDistribution):
    """
    The one-dimensional normal distribution. The normal distribution can be defined over a scalar random variable or an array of random variables. In case
    of an array of random variables, the mean and variance are broadcasted to the shape of the output random variable (array).

    :param mean: Mean of the normal distribution.
    :type mean: Variable
    :param variance: Variance of the normal distribution.
    :type variance: Variable
    :param rand_gen: the random generator (default: MXNetRandomGenerator).
    :type rand_gen: RandomGenerator
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, mean, variance, rand_gen=None, dtype=None, ctx=None):
        if not isinstance(mean, Variable):
            mean = Variable(value=mean)
        if not isinstance(variance, Variable):
            variance = Variable(value=variance)

        inputs = [('mean', mean), ('variance', variance)]
        input_names = [k for k, _ in inputs]
        output_names = ['random_variable']
        super(Normal, self).__init__(inputs=inputs, outputs=None,
                                     input_names=input_names,
                                     output_names=output_names,
                                     rand_gen=rand_gen, dtype=dtype, ctx=ctx)

    @UnivariateLogPDFDecorator()
    def log_pdf(self, mean, variance, random_variable, F=None):
        """
        Computes the logarithm of the probability density function (PDF) of the normal distribution.

        :param mean: the mean of the normal distribution.
        :type mean: MXNet NDArray or MXNet Symbol
        :param variance: the variance of the normal distributions.
        :type variance: MXNet NDArray or MXNet Symbol
        :param random_variable: the random variable of the normal distribution.
        :type random_variable: MXNet NDArray or MXNet Symbol
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: log pdf of the distribution.
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        F = get_default_MXNet_mode() if F is None else F
        logvar = np.log(2 * np.pi) / -2 + F.log(variance) / -2
        logL = F.broadcast_add(logvar, F.broadcast_div(F.square(
            F.broadcast_minus(random_variable, mean)), -2 * variance)) * self.log_pdf_scaling
        return logL

    @UnivariateDrawSamplesDecorator()
    def draw_samples(self, mean, variance, rv_shape, num_samples=1, F=None):
        """
        Draw samples from the normal distribution.

        :param mean: the mean of the normal distribution.
        :type mean: MXNet NDArray or MXNet Symbol
        :param variance: the variance of the normal distributions.
        :type variance: MXNet NDArray or MXNet Symbol
        :param rv_shape: the shape of each sample.
        :type rv_shape: tuple
        :param nSamples: the number of drawn samples (default: one).
        :int nSamples: int
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: a set samples of the normal distribution.
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        F = get_default_MXNet_mode() if F is None else F
        out_shape = (num_samples,) + rv_shape
        return F.broadcast_add(F.broadcast_mul(self._rand_gen.sample_normal(
            shape=out_shape, dtype=self.dtype, ctx=self.ctx),
            F.sqrt(variance)), mean)

    @staticmethod
    def define_variable(mean=0., variance=1., shape=None, rand_gen=None,
                        dtype=None, ctx=None):
        """
        Creates and returns a random variable drawn from a normal distribution.

        :param mean: Mean of the distribution.
        :param variance: Variance of the distribution.
        :param shape: the shape of the random variable(s).
        :type shape: tuple or [tuple]
        :param rand_gen: the random generator (default: MXNetRandomGenerator).
        :type rand_gen: RandomGenerator
        :param dtype: the data type for float point numbers.
        :type dtype: numpy.float32 or numpy.float64
        :param ctx: the mxnet context (default: None/current context).
        :type ctx: None or mxnet.cpu or mxnet.gpu
        :returns: the random variables drawn from the normal distribution.
        :rtypes: Variable
        """
        normal = Normal(mean=mean, variance=variance, rand_gen=rand_gen,
                        dtype=dtype, ctx=ctx)
        normal._generate_outputs(shape=shape)
        return normal.random_variable


class MultivariateNormalLogPDFDecorator(LogPDFDecorator):

    def _wrap_log_pdf_with_broadcast(self, func):
        def log_pdf_broadcast(self, F, **kw):
            """
            Computes the logrithm of the probability density/mass function (PDF/PMF) of the distribution. The inputs and outputs variables are in RTVariable format.

            Shape assumptions:
            * mean is S x N x D
            * covariance is S x N x D x D
            * random_variable is S x N x D

            Where:
            * S, the number of samples, is optional. If more than one of the variables has samples, the number of samples in each variable must be the same. S is 1 by default if not a sampled variable.
            * N is the number of data points. N can be any number of dimensions (N_1, N_2, ...) but must be broadcastable to the shape of random_variable.
            * D is the dimension of the distribution.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param kw: the dict of input and output variables of the distribution
            :type kw: {str (name): MXNet NDArray or MXNet Symbol}
            :returns: log pdf of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol
            """
            variables = {name: kw[name] for name, _ in self.inputs}
            variables['random_variable'] = kw['random_variable']
            rv_shape = variables['random_variable'].shape[1:]

            nSamples = max([get_num_samples(F, v) for v in variables.values()])

            shapes_map = {}
            shapes_map['mean'] = (nSamples,) + rv_shape
            shapes_map['covariance'] = (nSamples,) + rv_shape + (rv_shape[-1],)
            shapes_map['random_variable'] = (nSamples,) + rv_shape
            variables = {name: broadcast_to_w_samples(F, v, shapes_map[name])
                         for name, v in variables.items()}
            res = func(self, F=F, **variables)
            return res
        return log_pdf_broadcast


class MultivariateNormalDrawSamplesDecorator(DrawSamplesDecorator):

    def _wrap_draw_samples_with_broadcast(self, func):
        def draw_samples_broadcast(self, F, rv_shape, num_samples=1,
                                   always_return_tuple=False, **kw):
            """
            Draw a number of samples from the distribution. The inputs and outputs variables are in RTVariable format.

            :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray)
            :param rv_shape: the shape of each sample
            :type rv_shape: tuple
            :param num_samples: the number of drawn samples (default: one)
            :int num_samples: int
            :param always_return_tuple: Whether return a tuple even if there is only one variables in outputs.
            :type always_return_tuple: boolean
            :param kw: the dict of input variables of the distribution
            :type kw: {name: MXNet NDArray or MXNet Symbol}
            :returns: a set samples of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol or [MXNet NDArray or MXNet Symbol]
            """
            rv_shape = list(rv_shape.values())[0]
            variables = {name: kw[name] for name, _ in self.inputs}

            num_samples = max([get_num_samples(F, v) for v in variables.values()])
            shapes_map = {}
            shapes_map['mean'] = (num_samples,) + rv_shape
            shapes_map['covariance'] = (num_samples,) + rv_shape + (rv_shape[-1],)
            shapes_map['random_variable'] = (num_samples,) + rv_shape
            variables = {name: broadcast_to_w_samples(F, v, shapes_map[name])
                         for name, v in variables.items()}

            res = func(self, F=F, rv_shape=rv_shape, num_samples=num_samples,
                       **variables)
            if always_return_tuple:
                res = (res,)
            return res
        return draw_samples_broadcast


class MultivariateNormal(Distribution):
    """
    The multi-dimensional normal distribution.

    :param mean: Mean of the normal distribution.
    :type mean: Variable
    :param covariance: Covariance matrix of the distribution.
    :type covariance: Variable
    :param rand_gen: the random generator (default: MXNetRandomGenerator).
    :type rand_gen: RandomGenerator
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, mean, covariance, rand_gen=None, minibatch_ratio=1.,
                 dtype=None, ctx=None):
        self.minibatch_ratio = minibatch_ratio
        if not isinstance(mean, Variable):
            mean = Variable(value=mean)
        if not isinstance(covariance, Variable):
            covariance = Variable(value=covariance)

        inputs = [('mean', mean), ('covariance', covariance)]
        input_names = ['mean', 'covariance']
        output_names = ['random_variable']
        super(MultivariateNormal, self).__init__(inputs=inputs, outputs=None,
                                     input_names=input_names,
                                     output_names=output_names,
                                     rand_gen=rand_gen, dtype=dtype, ctx=ctx)

    def replicate_self(self, attribute_map=None):
        """
        Replicates this Factor, using new inputs, outputs, and a new uuid.
        Used during model replication to functionally replicate a factor into a new graph.

        :param inputs: new input variables of the factor.
        :type inputs: a dict of {'name' : Variable} or None
        :param outputs: new output variables of the factor.
        :type outputs: a dict of {'name' : Variable} or None
        """
        replicant = super(MultivariateNormal, self).replicate_self(attribute_map)
        replicant.minibatch_ratio = self.minibatch_ratio
        return replicant

    @MultivariateNormalLogPDFDecorator()
    def log_pdf(self, mean, covariance, random_variable, F=None):
        """
        Computes the logarithm of the probability density function (PDF) of the normal distribution.

        :param mean: the mean of the normal distribution.
        :type mean: MXNet NDArray or MXNet Symbol
        :param covariance: the covariance of the distribution.
        :type covariance: MXNet NDArray or MXNet Symbol
        :param random_variable: the random variable of the normal distribution.
        :type random_variable: MXNet NDArray or MXNet Symbol
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: log pdf of the distribution.
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        F = get_default_MXNet_mode() if F is None else F
        N = mean.shape[-1]
        lmat = F.linalg.potrf(covariance)
        logdetl = - F.linalg.sumlogdiag(F.abs(lmat)) # maybe sum if n x d x d
        targets = random_variable - mean
        zvec = F.sum(F.linalg.trsm(lmat, F.expand_dims(targets, axis=-1)), axis=-1)
        sqnorm_z = - F.sum(F.square(zvec), axis=-1)
        return 0.5 * (sqnorm_z - (N * np.log(2 * np.pi))) + logdetl

    @MultivariateNormalDrawSamplesDecorator()
    def draw_samples(self, mean, covariance, rv_shape, num_samples=1, F=None):
        """
        Draw a number of samples from the normal distribution.

        :param mean: the mean of the normal distribution.
        :type mean: MXNet NDArray or MXNet Symbol
        :param covariance: the covariance of the normal distributions.
        :type covariance: MXNet NDArray or MXNet Symbol
        :param rv_shape: the shape of each sample.
        :type rv_shape: tuple
        :param num_samples: the number of drawn samples (default: one).
        :int num_samples: int
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: a set samples of the normal distribution
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        F = get_default_MXNet_mode() if F is None else F
        out_shape = (num_samples,) + rv_shape + (1,)
        lmat = F.linalg.potrf(covariance)
        epsilon = self._rand_gen.sample_normal(
            shape=out_shape, dtype=self.dtype, ctx=self.ctx)
        lmat_eps = F.linalg.trmm(lmat, epsilon)
        return F.broadcast_add(lmat_eps.sum(-1), mean)

    @staticmethod
    def define_variable(shape, mean=0., covariance=None, rand_gen=None,
                        minibatch_ratio=1., dtype=None, ctx=None):
        """
        Creates and returns a random variable drawn from a normal distribution.

        :param mean: Mean of the distribution.
        :param covariance: Variance of the distribution.
        :param shape: the shape of the random variable(s).
        :type shape: tuple or [tuple]
        :param rand_gen: the random generator (default: MXNetRandomGenerator).
        :type rand_gen: RandomGenerator
        :param dtype: the data type for float point numbers.
        :type dtype: numpy.float32 or numpy.float64
        :param ctx: the mxnet context (default: None/current context).
        :type ctx: None or mxnet.cpu or mxnet.gpu
        :returns: the random variables drawn from the normal distribution.
        :rtypes: Variable
        """
        covariance = covariance if covariance is not None else mx.nd.array(np.eye(N=shape[-1]), dtype=dtype, ctx=ctx)
        normal = MultivariateNormal(mean=mean, covariance=covariance,
                                    rand_gen=rand_gen,
                                    dtype=dtype, ctx=ctx)
        normal._generate_outputs(shape=shape)
        return normal.random_variable

    def _generate_outputs(self, shape):
        """
        Set the output variable of the distribution.

        :param shape: the shape of the random distribution.
        :type shape: tuple
        """
        self.outputs = [('random_variable', Variable(value=self, shape=shape))]



import json
from ..components import ModelComponent
from ..common.exceptions import SerializationError


__GRAPH_JSON_VERSION__ = '1.0'


class ModelComponentEncoder(json.JSONEncoder):

    def default(self, obj):
        """
        Serializes a ModelComponent object. Note: does not serialize the successor attribute as it isn't  necessary for serialization.
        """
        import mxfusion.components as mf
        if isinstance(obj, mf.ModelComponent):
            return {
                "type": obj.__class__.__name__,
                "uuid": obj.uuid,
                "name": obj.name,
                "inherited_name": obj.inherited_name if hasattr(obj, 'inherited_name') else None,
                "attributes": [a.uuid for a in obj.attributes],
                "version": __GRAPH_JSON_VERSION__
            }
        return super(ModelComponentEncoder, self).default(obj)


class ModelComponentDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(
            self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        """
        Reloads a ModelComponent object. Note: does not reload the successor attribute as it isn't necessary for serialization.
        """
        if not isinstance(obj, type({})) or 'uuid' not in obj:
            return obj
        if obj['version'] != __GRAPH_JSON_VERSION__:
            raise SerializationError('The format of the stored model component '+str(obj['name'])+' is from an old version '+str(obj['version'])+'. The current version is '+__GRAPH_JSON_VERSION__+'. Backward compatibility is not supported yet.')
        v = ModelComponent()
        v._uuid = obj['uuid']
        v.attributes = obj['attributes']
        v.name = obj['name']
        v.inherited_name = obj['inherited_name']
        v.type = obj['type']
        return v



from .....common.config import get_default_dtype
from .....common.exceptions import ModelSpecificationError
from .....util.util import rename_duplicate_names, slice_axis
from ....variables import Variable

# TODO: write the design doc for kernels


class Kernel(object):
    """
    The base class for a Gaussian process kernel: a positive definite function which forms of a covariance function (kernel).

    :param input_dim: the number of dimensions of the kernel. (The total number of active dimensions).
    :type input_dim: int
    :param name: the name of the kernel. The name is also used as the prefix for the kernel parameters.
    :type name: str
    :param active_dims: The dimensions of the inputs that are taken for the covariance matrix computation. (default: None, taking all the dimensions).
    :type active_dims: [int] or None
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, input_dim, name, active_dims=None, dtype=None,
                 ctx=None):
        self.input_dim = input_dim
        self.name = name
        self.dtype = get_default_dtype() if dtype is None else dtype
        self.ctx = ctx
        self.active_dims = active_dims
        self._parameter_names = []

    def __setattr__(self, name, value):
        """
        Override to maintain a list of kernel parameters.
        """
        if isinstance(value, Variable):
            self._parameter_names.append(name)
        super(Kernel, self).__setattr__(name, value)

    @property
    def local_parameters(self):
        """
        The kernel parameters in the current kernel, which does not include kernel parameters that belongs to the sub-kernels of a compositional
        kernel. The keys of the returned dictionary are the name of the kernel parameters (without the prefix) and the values are the corresponding
        variables.

        :return: a dictionary of local kernel parameters, in which the keys are the name of individual parameters, including the kernel in front, and
            the values are the corresponding Variables.
        :rtype: {str: Variable}
        """
        return {getattr(self, n) for n in self._parameter_names}

    @property
    def parameters(self):
        """
        All the kernel parameters including the kernel parameters that belongs to the sub-kernels. The keys of the returned dictionary are the name of
        the kernel parameters with a prefix (the name of the kernel plus '_') and the values are the corresponding variables.

        :return: a dictionary of all the kernel parameters, in which the keys are the name of individual parameters, including the kernel in front,
            and the values are the corresponding Variables.
        :rtype: {str: Variable}
        """
        raise NotImplementedError

    def K(self, F, X, X2=None, **kernel_params):
        """
        Compute the covariance matrix.

        .. math::
            K_{ij} = k(X_i, X2_j)

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type \**kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :return: The covariance matrix
        :rtype: MXNet NDArray or MXNet Symbol
        """
        # Remove the prefix in the names of kernel_params
        # The prefix is defined as the kernel name plus _
        offset = len(self.name) + 1
        params = {k[offset:]: v for k, v in kernel_params.items() if
                  k.startswith(self.name + '_')}
        if self.active_dims is not None:
            X = slice_axis(F, X, axis=-1, indices=self.active_dims)
            if X2 is not None:
                X2 = slice_axis(F, X2, axis=-1, indices=self.active_dims)
        return self._compute_K(F=F, X=X, X2=X2, **params)

    def Kdiag(self, F, X, **kernel_params):
        """
        Compute the diagonal of the covariance matrix.

        .. math::
            K_{ii} = k(X_i, X_i)

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type \**kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :return: The diagonal of the covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        # Remove the prefix in the names of kernel_params
        # The prefix is defined as the kernel name plus _
        offset = len(self.name) + 1
        params = {k[offset:]: v for k, v in kernel_params.items() if
                  k.startswith(self.name + '_')}
        if self.active_dims is not None:
            X = slice_axis(F, X, axis=-1, indices=self.active_dims)
        return self._compute_Kdiag(F=F, X=X, **params)

    def add(self, other, name='sum'):
        """
        Construct a new kernel by adding this kernel to another kernel.

        :param other: the other kernel to be added.
        :type other: Kernel
        :return: the kernel which is the sum of the current kernel with the specified kernel.
        :rtype: Kernel
        """
        if not isinstance(other, Kernel):
            raise ModelSpecificationError("Only a Gaussian Process Kernel can be added to a Gaussian Process Kernel.")
        from .add_kernel import AddKernel
        return AddKernel([self, other], name=name, ctx=self.ctx,
                         dtype=self.dtype)

    def __add__(self, other):
        """
        Overwrite the "+" operator to perform summation of kernels.
        """
        return self.add(other)

    def _compute_K(self, F, X, X2=None, **kernel_params):
        """
        The internal interface for the actual covariance matrix computation.

        This function takes as an assumption: The prefix in the keys of *kernel_params* that corresponds to the name of the kernel has been
        removed. The dimensions of *X* and *X2* have been sliced according to *active_dims*.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type \**kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        raise NotImplementedError

    def _compute_Kdiag(self, F, X, **kernel_params):
        """
        The internal interface for the actual computation for the diagonal of the covariance matrix.

        This function takes as an assumption: The prefix in the keys of *kernel_params* that corresponds to the name of the kernel has been
        removed. The dimensions of *X* has been sliced according to *active_dims*.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type \**kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        raise NotImplementedError

    def fetch_parameters(self, params):
        """
        The helper function to fetch the kernel parameters from a set of variables according to the UUIDs of the kernel parameters. It returns a
        dictionary of kernel parameters, where the keys are the name of the kernel parameters and the values are the MXNet array at runtime. The
        returned dict can be directly passed into *K* and *Kdiag* as *kernel_params*.

        :param params: the set of parameters where the kernel parameters are fetched from.
        :type params: {str (UUID): MXNet NDArray or MXNet Symbol}
        :return: a dict of the kernel parameters, where the keys are the name of the kernel parameters and the values are the MXNet array at runtime.
        :rtype: {str (kernel name): MXNet NDArray or MXNet Symbol}
        """
        return {n: params[v.uuid] for n, v in self.parameters.items()}


class NativeKernel(Kernel):
    """
    The base class for all the native kernels: the computation of the covariance matrix does not depend on other kernels.

    :param input_dim: the number of dimensions of the kernel. (The total number of active dimensions).
    :type input_dim: int
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param active_dims: The dimensions of the inputs that are taken for the covariance matrix computation. (default: None, taking all the dimensions).
    :type active_dims: [int] or None
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, input_dim, name, active_dims=None, dtype=None,
                 ctx=None):
        super(NativeKernel, self).__init__(input_dim=input_dim, name=name,
                                           active_dims=active_dims,
                                           dtype=dtype, ctx=ctx)

    @property
    def parameters(self):
        """
        All the kernel parameters including the kernel parameters that belongs to the sub-kernels. The keys of the returned dictionary are the name of
        the kernel parameters with a prefix (the name of the kernel plus '_') and the values are the corresponding variables.

        :return: a dictionary of all the kernel parameters, in which the keys are the name of individual parameters, including the kernel in front,
            and the values are the corresponding Variables.
        :rtype: {str: Variable}
        """
        return {self.name + '_' + n: getattr(self, n) for n in
                self._parameter_names}


class CombinationKernel(Kernel):
    """
    The base class for combination kernels: the covariance matrix is computed by combining the covariance matrix from multiple sub-kernels.

    :param sub_kernels: a list of kernels that are combined to compute a covariance matrix.
    :type sub_kernels: [Kernel]
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, sub_kernels, name, dtype=None, ctx=None):
        input_dim = max([k.input_dim for k in sub_kernels])
        renaming = rename_duplicate_names([k.name for k in sub_kernels])
        for i, n in renaming:
            sub_kernels[i].name = n
        super(CombinationKernel, self).__init__(
            input_dim=input_dim, name=name, dtype=dtype, ctx=ctx)
        self.sub_kernels = sub_kernels
        for k in self.sub_kernels:
            setattr(self, k.name, k)

    @property
    def parameters(self):
        """
        All the kernel parameters including the kernel parameters that belongs to the sub-kernels. The keys of the returned dictionary are the name of
        the kernel parameters with a prefix (the name of the kernel plus '_') and the values are the corresponding variables.

        :return: a dictionary of all the kernel parameters, in which the keys are the name of individual parameters, including the kernel in front,
            and the values are the corresponding Variables.
        :rtype: {str: Variable}
        """
        p = {}
        for k in self.sub_kernels:
            p.update(k.parameters)
        return {self.name + '_' + k: v for k, v in p.items()}



from ..models.posterior import Posterior
from ..components.variables import PositiveTransformation
from ..components.variables import Variable, VariableType
from ..components.distributions.normal import Normal
from ..util.inference import variables_to_UUID


def create_Gaussian_meanfield(model, observed):
    """
    Create the Meanfield posterior for Variational Inference.

    :param model_graph: the definition of the probabilistic model
    :type model_graph: Model
    :param observed: A list of observed variables
    :type observed: [Variable]
    :returns: the resulting posterior representation
    :rtype: Posterior
    """
    observed = variables_to_UUID(observed)
    q = Posterior(model)
    for v in model.variables.values():
        if v.type == VariableType.RANDVAR and v not in observed:
            mean = Variable(shape=v.shape)
            variance = Variable(shape=v.shape,
                                transformation=PositiveTransformation())
            q[v].set_prior(Normal(mean=mean, variance=variance))
    return q



from .kernel import CombinationKernel


class AddKernel(CombinationKernel):
    """
    The add kernel that computes a covariance matrix by suming the covariance
    matrices of a list of kernels.

    :param sub_kernels: a list of kernels that are combined to compute a covariance matrix.
    :type sub_kernels: [Kernel]
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, sub_kernels, name='add', dtype=None, ctx=None):
        kernels = []
        for k in sub_kernels:
            if isinstance(k, AddKernel):
                for k2 in k.sub_kernels:
                    kernels.append(k2)
            else:
                kernels.append(k)
        super(AddKernel, self).__init__(
            sub_kernels=kernels, name=name, dtype=dtype, ctx=ctx)

    def _compute_K(self, F, X, X2=None, **kernel_params):
        """
        The internal interface for the actual covariance matrix computation.

        This function takes as an assumption: The prefix in the keys of
        *kernel_params* that corresponds to the name of the kernel has been
        removed. The dimensions of *X* and *X2* have been sliced according to
        *active_dims*.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type **kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        K = self.sub_kernels[0].K(F=F, X=X, X2=X2, **kernel_params)
        for k in self.sub_kernels[1:]:
            K = K + k.K(F=F, X=X, X2=X2, **kernel_params)
        return K

    def _compute_Kdiag(self, F, X, **kernel_params):
        """
        The internal interface for the actual computation for the diagonal of the covariance matrix.

        This function takes as an assumption: The prefix in the keys of *kernel_params* that corresponds to the name of the kernel has been
        removed. The dimensions of *X* has been sliced according to *active_dims*.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param \**kernel_params: the set of kernel parameters, provided as keyword arguments.
        :type \**kernel_params: {str: MXNet NDArray or MXNet Symbol}
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        K = self.sub_kernels[0].Kdiag(F=F, X=X, **kernel_params)
        for k in self.sub_kernels[1:]:
            K = K + k.Kdiag(F=F, X=X, **kernel_params)
        return K



from ..common.exceptions import ModelSpecificationError
from ..components.variables import Variable, VariableType


def variables_to_UUID(variables):
    return [v.uuid if isinstance(v, Variable) else v for v in variables]


def realize_shape(shape, constants):
    rshape = []
    for s in shape:
        if isinstance(s, int):
            rshape.append(s)
        elif isinstance(s, Variable):
            if s.type == VariableType.CONSTANT:
                rshape.append(s.get_constant())
            elif s.type == VariableType.PARAMETER:
                rshape.append(constants[s.uuid])
    return tuple(rshape)


def discover_shape_constants(data_shapes, graphs):
    """
    The internal shape discovery from data and the shape inference for all the
    variables in the model and inference models.

    :param data_shapes: a dict of shapes of data
    :param graphs: a list of factor graphs of which variable shapes are
        searched.
    :returns: a dict of constants discovered from data shapes
    :rtype: {Variable: int}
    """
    shape_constants = {}
    variables = {}
    for g in graphs:
        variables.update(g.variables)
    for var_id, shape in data_shapes.items():
        def_shape = variables[var_id].shape
        for s1, s2 in zip(def_shape, shape):
            if isinstance(s1, int):
                if s1 != s2:
                    raise ModelSpecificationError("Variable shape mismatch! s1 : {} s2 : {}".format(str(s1), str(s2)))
            elif isinstance(s1, Variable):
                shape_constants[s1] = s2
            else:
                raise ModelSpecificationError("The shape of a Variable should either an integer or a Variable, but encountered {}!".format(str(type(s1))))
    return shape_constants


def init_outcomes(inference_outcomes):
    if isinstance(inference_outcomes, list):
        updated_outcomes = inference_outcomes
    elif inference_outcomes is not None:
        updated_outcomes = [inference_outcomes]
    else:
        updated_outcomes = []
    return updated_outcomes



"""This module contains implementations of Gaussian Processes for MXFusion.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    gp
    cond_gp
    kernels
"""

__all__ = ['kernels', 'gp', 'cond_gp']

from .gp import GaussianProcess
from .cond_gp import ConditionalGaussianProcess



"""Contains the Variable class, Variable transformations, and runtime methods on variables.

Submodules
==========

.. autosummary::
    :toctree: _autosummary

    runtime_variable
    var_trans
    variable
"""

__all__ = ['runtime_variable', 'var_trans', 'variable']

from .runtime_variable import add_sample_dimension, add_sample_dimension_arrays, expectation, is_sampled_array, get_num_samples, as_samples
from .var_trans import Softplus, PositiveTransformation
from .variable import Variable, VariableType



from ..factor import Factor
from .random_gen import MXNetRandomGenerator
from ...util.inference import realize_shape
from ...common.config import get_default_dtype


class LogPDFDecorator(object):
    """
    The decorator for the log_pdf function in Distribution
    """
    def __call__(self, func):

        func_reshaped = self._wrap_log_pdf_with_broadcast(func)
        func_variables = self._wrap_log_pdf_with_variables(func_reshaped)
        return func_variables

    def _wrap_log_pdf_with_variables(self, func):

        def log_pdf_variables(self, F, variables):
            """
            Computes the logrithm of the probability density/mass function
            (PDF/PMF) of the distribution. The inputs and outputs variables are
            fetched from the *variables* argument according to their UUIDs.

            :param F: the MXNet computation mode
            :type F: mxnet.symbol or mxnet.ndarray
            :param variables: the set of MXNet arrays that holds the values of
            variables at runtime.
            :type variables: {str(UUID): MXNet NDArray or MXNet Symbol}
            :returns: log pdf of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol
            """
            args = {}
            for name, var in self.inputs:
                args[name] = variables[var.uuid]
            for name, var in self.outputs:
                args[name] = variables[var.uuid]
            return func(self, F=F, **args)
        return log_pdf_variables

    def _wrap_log_pdf_with_broadcast(self, func):
        raise NotImplementedError


class DrawSamplesDecorator(object):
    """
    The decorator for the draw_samples function in Distribution
    """
    def __call__(self, func):

        func_reshaped = self._wrap_draw_samples_with_broadcast(func)
        func_variables = self._wrap_draw_samples_with_variables(func_reshaped)
        return func_variables

    def _wrap_draw_samples_with_variables(self, func):

        def draw_samples_variables(self, F, variables, num_samples=1,
                                   always_return_tuple=False):
            """
            Draw a set of samples from the distribution. The inputs variables
            are fetched from the *variables* argument according to their UUIDs.

            :param F: the MXNet computation mode
            :type F: mxnet.symbol or mxnet.ndarray
            :param variables: the set of MXNet arrays that holds the values of
            variables at runtime.
            :type variables: {str(UUID): MXNet NDArray or MXNet Symbol}
            :param num_samples: the number of drawn samples (default: one)
            :int num_samples: int
            :param always_return_tuple: Whether return a tuple even if there is
            only one variables in outputs.
            :type always_return_tuple: boolean
            :returns: a set samples of the distribution
            :rtypes: MXNet NDArray or MXNet Symbol or [MXNet NDArray or MXNet
            Symbol]
            """
            args = {}
            for name, var in self.inputs:
                args[name] = variables[var.uuid]
            args['rv_shape'] = {name: realize_shape(rv.shape, variables) for
                                name, rv in self.outputs}
            return func(self, F=F, num_samples=num_samples,
                        always_return_tuple=always_return_tuple, **args)
        return draw_samples_variables

    def _wrap_draw_samples_with_broadcast(self, func):
        raise NotImplementedError


class Distribution(Factor):
    """
    The base class of a probability distribution associated with one or a set of random variables.

    :param inputs: the input variables that parameterize the probability distribution.
    :type inputs: {name: Variable}
    :param outputs: the random variables drawn from the distribution.
    :type outputs: {name: Variable}
    :param rand_gen: the random generator (default: MXNetRandomGenerator).
    :type rand_gen: RandomGenerator
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, inputs, outputs, input_names, output_names, rand_gen=None, dtype=None, ctx=None):
        super(Distribution, self).__init__(inputs=inputs, outputs=outputs,
                                           input_names=input_names,
                                           output_names=output_names)
        self._rand_gen = MXNetRandomGenerator if rand_gen is None else \
            rand_gen
        self.dtype = get_default_dtype() if dtype is None else dtype
        self.ctx = ctx
        self.log_pdf_scaling = 1

    def replicate_self(self, attribute_map=None):
        """
        Replicates this Factor, using new inputs, outputs, and a new uuid.
        Used during model replication to functionally replicate a factor into a
        new graph.

        :param inputs: new input variables of the factor.
        :type inputs: a dict of {'name' : Variable} or None
        :param outputs: new output variables of the factor.
        :type outputs: a dict of {'name' : Variable} or None
        """
        replicant = super(Distribution, self).replicate_self(attribute_map)
        replicant._rand_gen = self._rand_gen
        replicant.dtype = self.dtype
        replicant.ctx = self.ctx
        return replicant

    def log_pdf(self, F=None, **kwargs):
        """
        Computes the logarithm of the probability density/mass function (PDF/PMF) of the distribution.

        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: log pdf of the distribution
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        raise NotImplementedError

    def log_cdf(self, F=None, **kwargs):
        """
        Computes the logarithm of the cumulative distribution function (CDF) of the distribution.

        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: log cdf of the distribution.
        :rtypes: MXNet NDArray or MXNet Symbol
        """
        raise NotImplementedError

    def draw_samples(self, rv_shape, num_samples=1, F=None, **kwargs):
        """
        Draw a number of samples from the distribution.

        :param rv_shape: the shape of each sample.
        :type rv_shape: tuple, [tuple]
        :param num_samples: the number of drawn samples (default: one).
        :int num_samples: int
        :param F: the MXNet computation mode (mxnet.symbol or mxnet.ndarray).
        :returns: a set samples of the distribution.
        :rtypes: MXNet NDArray or MXNet Symbol or [MXNet NDArray or MXNet Symbol]
        """
        raise NotImplementedError

    @staticmethod
    def define_variable(shape=None, rand_gen=None, dtype=None, ctx=None, **kwargs):
        """
        Define a random variable that follows from the specified distribution.

        :param shape: the shape of the random variable(s).
        :type shape: tuple or [tuple]
        :param rand_gen: the random generator (default: MXNetRandomGenerator).
        :type rand_gen: RandomGenerator
        :param dtype: the data type for float point numbers.
        :type dtype: numpy.float32 or numpy.float64
        :param ctx: the mxnet context (default: None/current context).
        :type ctx: None or mxnet.cpu or mxnet.gpu
        :param kwargs: the input variables that parameterize the probability
        distribution.
        :type kwargs: {name: Variable}
        :returns: the random variables drawn from the distribution.
        :rtypes: Variable or [Variable]
        """
        raise NotImplementedError



import os
from setuptools import setup, find_packages
from mxfusion.__version__ import __version__

with open('README.md', 'r') as fh:
    long_description = fh.read()

with open('requirements/requirements.txt', 'r') as req:
    requires = req.read().split("\n")

setup(
    name='MXFusion',  # this is the name of the package as you will import it i.e import package-name
    version=__version__,
    author='Eric Meissner',
    author_email='meissner.eric.7@gmail.com',
    description=' Modular Probabilistic Programming on MXNet',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/amzn/MXFusion',
    packages=find_packages(exclude=['testing*']),
    include_package_data=True,
    install_requires=requires,
    license='Apache License 2.0',
    classifiers=(
       # https://pypi.org/pypi?%3Aaction=list_classifiers
       'Development Status :: 5 - Production/Stable',
       'Intended Audience :: Developers',
       'Intended Audience :: Education',
       'Intended Audience :: Science/Research',
       'Programming Language :: Python :: 3',
       'Programming Language :: Python :: 3.4',
       'Programming Language :: Python :: 3.6',
       'Programming Language :: Python :: 3.7',
       'License :: OSI Approved :: Apache Software License',
       'Operating System :: OS Independent',
       'Topic :: Scientific/Engineering :: Artificial Intelligence',
       'Topic :: Scientific/Engineering :: Mathematics',
       'Topic :: Software Development :: Libraries',
       'Topic :: Software Development :: Libraries :: Python Modules'
    ),
)



from abc import ABC, abstractmethod


class GradLoop(ABC):
    """
    The abstract class for the loop of gradient-based inference.
    """

    @abstractmethod
    def run(self, infr_executor, data, param_dict, ctx, optimizer='adam',
            learning_rate=1e-3, max_iter=2000, verbose=False):
        """
        :param infr_executor: The MXNet function that computes the training objective.
        :type infr_executor: MXNet Gluon Block
        :param data: a list of observed variables
        :type data: [mxnet.ndarray]
        :param param_dict: The MXNet ParameterDict for Gradient-based optimization
        :type param_dict: mxnet.gluon.ParameterDict
        :param ctx: MXNet context
        :type ctx: mxnet.cpu or mxnet.gpu
        :param optimizer: the choice of optimizer (default: 'adam')
        :type optimizer: str
        :param learning_rate: the learning rate of the gradient optimizer (default: 0.001)
        :type learning_rate: float
        :param max_iter: the maximum number of iterations of gradient optimization
        :type max_iter: int
        :param verbose: whether to print per-iteration messages.
        :type verbose: boolean
        """
        pass



from abc import ABC
import mxnet as mx
from ...common.config import get_default_dtype, get_default_MXNet_mode


class RandomGenerator(ABC):
    """
    The abstract class of the pseudo-random number generator.
    """

    @staticmethod
    def sample_normal(loc=0, scale=1, shape=None, dtype=None, out=None,
                      ctx=None):
        pass


class MXNetRandomGenerator(RandomGenerator):
    """
    The MXNet pseudo-random number generator.
    """

    @staticmethod
    def sample_normal(loc=0, scale=1, shape=None, dtype=None, out=None,
                      ctx=None, F=None):
        """
        :param ref_var: reference variable for getting the context of execution.
        """
        dtype = get_default_dtype() if dtype is None else dtype
        F = get_default_MXNet_mode() if F is None else F
        if F is mx.ndarray:
            return F.random.normal(loc=loc, scale=scale, shape=shape,
                                   dtype=dtype, ctx=ctx, out=out)
        else:
            return F.random.normal(loc=loc, scale=scale, shape=shape,
                                   dtype=dtype, out=out)

    @staticmethod
    def sample_multinomial(data, get_prob=True, dtype='int32',
                           F=None):
        F = get_default_MXNet_mode() if F is None else F
        return F.random.multinomial(
            data=data, get_prob=get_prob, dtype=dtype)



from enum import Enum
import mxnet as mx
import numpy as np
from ...common.exceptions import ModelSpecificationError
from ..model_component import ModelComponent


class VariableType(Enum):

    """
    Variables types:

    * CONSTANT: a variable with a constant value
    * PARAMETER: a variable with no prior distribution
    * RANDVAR: a variable that follows a distribution or module
    * FUNCVAR: a variable that is an output of a function
    """

    CONSTANT = 0
    PARAMETER = 1
    RANDVAR = 2
    FUNCVAR = 3


class Variable(ModelComponent):
    """
    Define a variable in MXFusion. The variable can be either a constant, a parameter, an outcome of a function evaluation or a random variable
    following a probabilistic distribution.

    :param value: The value of variable. If it is a numpy or MXNet array, the variable is considered as a constant. If it is a function evaluation, the
        varaible is considered as the outcome of a function evaluation. If it is a probabilitistic distribution, the variable is considered as a random
        variable. If it is None, the variable is considered as a parameter.
    :type value: (optional) None or numpy array or MXNet array or float or int or FunctionEvaluation or Distribution.
    :param shape: The expected shape of the Variable.
    :type shape: (optional) tuple or None
    :param transformation: The constraint on the domain of the variable, e.g., a positive value.
    :type transformation: (optional) transformation or None
    :param isInherited: True if the Variable is constructed from a GluonBlock parameter.
    :type isInherited: bool
    """
    def __init__(self, value=None, shape=None, transformation=None, isInherited=False, initial_value=None):
        super(Variable, self).__init__()

        self.shape = shape if shape is not None else (1,)
        self.attributes = [s for s in self.shape if isinstance(s, Variable)]
        # whether the variable is inherited from a Gluon block.
        self.isInherited = isInherited
        self._transformation = transformation
        self._value = None
        self._initial_value = initial_value
        self.isConstant = False
        from ..distributions import Distribution
        from ..functions.function_evaluation import FunctionEvaluation
        if isinstance(value, Distribution):
            self._initialize_as_randvar(value, shape, transformation)
        elif isinstance(value, FunctionEvaluation):
            self._initialize_as_funcvar(value, shape, transformation)
        else:
            self._initialize_as_param(value, shape, transformation)

    @property
    def type(self):
        from ..distributions import Distribution
        from ..functions import FunctionEvaluation
        if self.factor is None:
            if self.isConstant:
                return VariableType.CONSTANT
            else:
                return VariableType.PARAMETER
        elif isinstance(self.factor, Distribution):
            return VariableType.RANDVAR
        elif isinstance(self.factor, FunctionEvaluation):
            return VariableType.FUNCVAR

    def replicate_self(self, attribute_map=None):
        """
        This functions as a copy constructor for the object. In order to do a copy constructor we first call ``__new__`` on the class which creates a blank object.
        We then initialize that object using the methods standard init procedures, and do any extra copying of attributes.

        Replicates this Factor, using new inputs, outputs, and a new uuid. Used during model replication to functionally replicate a factor into a new graph.

        :param attribute_map: A mapping from attributes of this object that were Variables to thier replicants.
        :type attribute_map: {Variable: replicated Variable}
        """
        if attribute_map is not None:
            shape = tuple(attribute_map[s] if isinstance(s, Variable) else s for s in self.shape)
        else:
            shape = self.shape
        self.attributes = [s for s in shape if isinstance(s, Variable)]
        if self.type == VariableType.RANDVAR or self.type == \
           VariableType.FUNCVAR:
            v = Variable(value=None, shape=shape,
                         transformation=self.transformation)
        elif self.type == VariableType.PARAMETER:
            v = Variable(value=None, shape=shape,
                         transformation=self.transformation)
        elif self.type == VariableType.CONSTANT:
            v = Variable(value=self._value, shape=shape,
                         transformation=self.transformation)
        if self.isInherited:
            v.isInherited = self.isInherited
            v.inherited_name = self.inherited_name
        v._uuid = self.uuid
        # v.replicant = True
        return v

    @property
    def shape_str(self):
        ss = '('+', '.join([(str(i.type)[13:] if i.name is None else i.name) if isinstance(i, Variable) else str(i) for i in self.shape])+')'
        return ss

    def display_str(self, temp_name=None):
        name = temp_name if temp_name is not None else self.name
        name = "Variable(%s)" % self.uuid[:5] if name is None else name
        string = name
        return string

    def __repr__(self):
        return self.display_str()

    def _initialize_as_param(self, value, shape, transformation):

        if value is None:
            # Initialize as VariableType.PARAMETER
            if shape is None:
                self.shape = (1,)
        else:
            # Initialize as VariableType.CONSTANT
            self.isConstant = True
            if isinstance(value, np.ndarray):
                if shape is not None and shape != value.shape:
                    raise ModelSpecificationError("Shape mismatch in Variable creation. The numpy array shape " + str(value.shape) + " does not no match with the shape argument " + str(shape) + ".")
                value = mx.nd.array(value)
            elif isinstance(value, mx.nd.NDArray):
                if shape is not None and shape != value.shape:
                    raise ModelSpecificationError("Shape mismatch in Variable creation. The MXNet array shape " + str(value.shape) + " does not no match with the shape argument " + str(shape) + ".")
            elif isinstance(value, (float, int)):
                self.shape = (1,)
            self._value = value

    def _initialize_as_randvar(self, value, shape, transformation):
        if transformation is not None:
            raise NotImplementedError('Contraints on random variables are not supported!')

    def _initialize_as_funcvar(self, value, shape, transformation):
        self._inputs = [value]
        if shape is None:
            raise ModelSpecificationError("The shape argument was not given when defining a variable as the outcome of a function evaluation.")
        if transformation is not None:
            raise NotImplementedError('Contraints on function outputs are not supported!')

    def set_prior(self, distribution):
        """
        Set the prior over this variable.

        :param: distribution The distribution the variable is drawn from.
        """

        self.assign_factor(distribution)

    def assign_factor(self, factor):
        """
        Assign (or reassign) the distribution this variable is drawn from.

        :param: distribution The distribution this variable is drawn from.
        """
        factor.set_outputs(self)

    @property
    def factor(self):
        if self.predecessors is not None and len(self.predecessors) > 0:
            return self.predecessors[0][1]
        else:
            return None

    @property
    def constant(self):
        """
        Return the stored constant.

        :returns: the constant value
        """
        if self.type == VariableType.CONSTANT:
            return self._value
        else:
            raise ModelSpecificationError("The constant property is not accessible for variable with the type "+str(self.type)+".")

    @property
    def transformation(self):
        return self._transformation

    @property
    def initial_value(self):
        return self._initial_value



from ..variables import Variable
from .univariate import UnivariateDistribution, UnivariateLogPDFDecorator, UnivariateDrawSamplesDecorator
from ...util.customop import broadcast_to_w_samples


class PointMass(UnivariateDistribution):
    """
    The Point Mass distribution.

    :param value: the location of the point mass.
    """
    def __init__(self, location, rand_gen=None, dtype=None, ctx=None):
        inputs = [('location', location)]
        input_names = ['location']
        output_names = ['random_variable']
        super(PointMass, self).__init__(inputs=inputs, outputs=None,
                                        input_names=input_names,
                                        output_names=output_names,
                                        rand_gen=rand_gen, dtype=dtype, ctx=ctx)

    @UnivariateLogPDFDecorator()
    def log_pdf(self, location, random_variable, F=None):
        """
        Computes the logaorithm of probabilistic density function of the normal distribution.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param location: the location of the point mass.
        :param random_variable: the point to compute the logpdf for.
        :returns: An operator chain to compute the logpdf of the Normal distribution.
        """
        return 0.

    @UnivariateDrawSamplesDecorator()
    def draw_samples(self, location, rv_shape, num_samples=1, F=None):
        return broadcast_to_w_samples(F, location, False, (num_samples,) +
                                      rv_shape)

    @staticmethod
    def define_variable(location, shape=None, rand_gen=None, dtype=None,
                        ctx=None):
        """
        Creates and returns a random variable drawn from a Normal distribution.

        :param location: the location of the point mass.
        :param shape: Shape of random variables drawn from the distribution. If non-scalar, each variable is drawn iid.

        :returns: RandomVariable drawn from the distribution specified.
        """
        if not isinstance(location, Variable):
            loc = Variable(value=location)

        p = PointMass(location=loc, rand_gen=rand_gen, dtype=dtype, ctx=ctx)
        p._generate_outputs(shape=shape)
        return p.random_variable



from .kernel import NativeKernel
from ....variables import Variable
from ....variables import PositiveTransformation
from .....util.customop import broadcast_to_w_samples


class StationaryKernel(NativeKernel):
    """
    The base class for Stationary kernels (covariance functions).

    Stationary kernels (covariance functions).
    Stationary covariance fucntion depend only on r^2, where r^2 is defined as
    .. math::
        r2(x, x') = \\sum_{q=1}^Q (x_q - x'_q)^2
    The covariance function k(x, x' can then be written k(r).

    In this implementation, r is scaled by the lengthscales parameter(s):
    .. math::
        r2(x, x') = \\sum_{q=1}^Q \\frac{(x_q - x'_q)^2}{\\ell_q^2}.
    By default, there's only one lengthscale: seaprate lengthscales for each dimension can be enables by setting ARD=True.


    :param input_dim: the number of dimensions of the kernel. (The total number of active dimensions).
    :type input_dim: int
    :param ARD: a binary switch for Automatic Relevance Determination (ARD). If true, the squared distance is divided by a lengthscale for individual
        dimensions.
    :type ARD: boolean
    :param variance: the initial value for the variance parameter (scalar), which scales the whole covariance matrix.
    :type variance: float or MXNet NDArray
    :param lengthscale: the initial value for the lengthscale parameter.
    :type lengthscale: float or MXNet NDArray
    :param name: the name of the kernel. The name is used to access kernel parameters.
    :type name: str
    :param active_dims: The dimensions of the inputs that are taken for the covariance matrix computation. (default: None, taking all the dimensions).
    :type active_dims: [int] or None
    :param dtype: the data type for float point numbers.
    :type dtype: numpy.float32 or numpy.float64
    :param ctx: the mxnet context (default: None/current context).
    :type ctx: None or mxnet.cpu or mxnet.gpu
    """
    def __init__(self, input_dim, ARD=False, variance=1., lengthscale=1.,
                 name='stationary', active_dims=None, dtype=None, ctx=None):
        super(StationaryKernel, self).__init__(
            input_dim=input_dim, name=name, active_dims=active_dims,
            dtype=dtype, ctx=ctx)
        self.ARD = ARD
        if not isinstance(variance, Variable):
            variance = Variable(shape=(1,),
                                transformation=PositiveTransformation(),
                                initial_value=variance)
        if not isinstance(lengthscale, Variable):
            lengthscale = Variable(shape=(input_dim if ARD else 1,),
                                   transformation=PositiveTransformation(),
                                   initial_value=lengthscale)
        self.variance = variance
        self.lengthscale = lengthscale

    def _compute_R2(self, F, X, lengthscale, variance, X2=None):
        """
        The helper function that computes the squared distance for a stationary kernel.

        Compute the covariance matrix.

        .. math::
            r2(x, x') = \\sum_{q=1}^Q \\frac{(x_q - x'_q)^2}{\\ell_q^2}.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param X2: (optional) the second set of arguments to the kernel. If X2 is None, this computes a square covariance matrix of X. In other words,
            X2 is internally treated as X.
        :type X2: MXNet NDArray or MXNet Symbol
        :return: The squared distance.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        lengthscale = F.expand_dims(lengthscale, axis=-2)
        if X2 is None:
            xsc = X / broadcast_to_w_samples(F, lengthscale, X.shape)
            amat = F.linalg.syrk(xsc) * -2
            dg_a = F.sum(F.square(xsc), axis=-1)
            amat = F.broadcast_add(amat, F.expand_dims(dg_a, axis=-1))
            amat = F.broadcast_add(amat, F.expand_dims(dg_a, axis=-2))
        else:
            x1sc = X / broadcast_to_w_samples(F, lengthscale, X.shape)
            x2sc = X2 / broadcast_to_w_samples(F, lengthscale, X2.shape)
            amat = F.linalg.gemm2(x1sc, x2sc, False, True) * -2
            dg1 = F.sum(F.square(x1sc), axis=-1, keepdims=True)
            amat = F.broadcast_add(amat, dg1)
            dg2 = F.expand_dims(F.sum(F.square(x2sc), axis=-1), axis=-2)
            amat = F.broadcast_add(amat, dg2)
        return amat

    def _compute_Kdiag(self, F, X, lengthscale, variance):
        """
        The internal interface for the actual computation for the diagonal of the covariance matrix.

        :param F: MXNet computation type <mx.sym, mx.nd>.
        :param X: the first set of inputs to the kernel.
        :type X: MXNet NDArray or MXNet Symbol
        :param variance: the variance parameter (scalar), which scales the whole covariance matrix.
        :type variance: MXNet NDArray or MXNet Symbol
        :param lengthscale: the lengthscale parameter.
        :type lengthscale: MXNet NDArray or MXNet Symbol
        :return: The covariance matrix.
        :rtype: MXNet NDArray or MXNet Symbol
        """
        return broadcast_to_w_samples(F, variance, X.shape[:-1])
